import pandas as pd
import upsetplot
import gzip
from collections import Counter
from itertools import compress
from matplotlib import pyplot


def count_scores(pk_results, genomes, exclusive=False):
    """Compute score counts for input to upsetplot

    Parameters
    ----------
    pk_results : PKResults
        PKResults object representing superset index
    genomes
        iterable of genomes included in subset
    """

    genomes_ordered = tuple(g for g in pk_results.genomes if g in genomes)
    genomes_set = set(genomes)
    counts = {}
    for b, n in Counter(score for _, score in pk_results).items():
        expanded_score = pk_results.decode_score(b)
        if exclusive and bool(
            set(compress(pk_results.genomes, expanded_score)) - genomes_set
        ):
            continue
        membership = tuple(
            sorted(
                set(compress(pk_results.genomes, expanded_score)).intersection(
                    genomes_set
                )
            )
        )
        if membership:
            counts[membership] = counts.get(membership, 0) + n
    return pd.Series(
        counts.values(),
        index=pd.MultiIndex.from_tuples(
            (tuple(g in k for g in genomes_ordered) for k in counts.keys()),
            names=genomes_ordered,
        ),
    )


def upset(
    pk_results,
    output,
    genomes,
    vertical=False,
    show_counts=False,
    min_subset_size=None,
    max_subset_size=None,
    exclusive=False,
    table=None,
):
    score_counts = count_scores(pk_results, genomes, exclusive=exclusive)
    if table:
        with (gzip.open if table.endswith(".gz") else open)(table, "wb") as f:
            score_counts.to_csv(f, sep="\t", header=["k-mers"])
    upsetplot.plot(
        score_counts,
        orientation="vertical" if vertical else "horizontal",
        show_counts=show_counts,
        min_subset_size=min_subset_size,
        max_subset_size=max_subset_size,
    )
    pyplot.savefig(output)



import re
import os
import os.path

COORD_REGEX = re.compile(
    os.environ.get(
        "PANKMER_COORD_REGEX",  # Not used yet, but will be useful later
        "[\s\S]+:[0-9]+-[0-9]+$",
    )
)
# '([Cc]hr)?[0-9XY]+:[0-9]+-[0-9]+$'))
GENES_PATH = os.environ.get(
    "PANKMER_GENES_PATH"
)  # Not used yet, but will be useful later

EXAMPLE_DATA_URL = "https://salk-tm-pub.s3.us-west-2.amazonaws.com/PanKmer_example/PanKmer_example_Sp_Chr19.tar.gz"
EXAMPLE_DATA_DIR = os.environ.get("PANKMER_EXAMPLE_DATA_DIR", os.path.dirname(__file__))
KMER_SIZE = os.environ.get("PANKMER_KMER_SIZE", 31)



from os.path import isfile
import shutil
import tarfile
from pankmer.index import subset as _subset
from pankmer.version import __version__


def subset(index, output, genomes, exclusive=False, gzip_level=6):
    output_is_tar = output.endswith(".tar")
    output_dir = output[:-4] if output_is_tar else output
    _subset(
        str(index),
        str(index) if isfile(index) and tarfile.is_tarfile(index) else "",
        genomes,
        output_dir,
        gzip_level,
        exclusive,
    )
    if output_is_tar:
        with tarfile.open(output, "w") as tar:
            tar.add(output_dir)
        shutil.rmtree(output_dir)



from Bio.bgzf import BgzfWriter
import gff2bed
import json
import math
import os.path
import os
import pybedtools
import pysam
import subprocess

from pankmer.tree import adj_to_jaccard
from pankmer.adjacency_matrix import get_adjacency_matrix
from pankmer.anchor import check_for_bgzip, get_chromosome_sizes_from_anchor

LOWRES_STEP = 100


def anchormap(pk_results, output_dir, anchors, anno, threads: int = 1):
    """Generate an anchormap for visualization

    Parameters
    ----------
    pk_results : PKResults
        a PKResults object
    output_dir
        directory for output files
    anchors
        iterable of paths to anchor genomes
    anno
        iterable of paths to annotation files
    threads : int
        number of threads to use
    """
    for anchor in anchors:
        check_for_bgzip(anchor)
    os.mkdir(output_dir)
    os.mkdir(os.path.join(output_dir, "anchors"))
    os.mkdir(os.path.join(output_dir, "anno"))
    print("loading index")
    pk_results.threads = threads
    print("writing distance matrix")
    (1 - adj_to_jaccard(get_adjacency_matrix(pk_results))).to_csv(
        os.path.join(output_dir, 'distance.tsv'), sep='\t')
    score_bitsize = math.ceil(pk_results.number_of_genomes/8)
    print('writing config')
    sizes = {'.'.join(os.path.basename(a).split('.')[:-2]):
             get_chromosome_sizes_from_anchor(a).set_index('name')['size']
             for a in anchors}
    with open(os.path.join(output_dir,'config.json'), 'w') as f:
        json.dump({
            'prefix': output_dir,
            'genomes': list(pk_results.genomes),
            'anchors': ['.'.join(os.path.basename(a).split('.')[:-2])
                        for a in anchors],
            'lowres_step': LOWRES_STEP,
            'sizes': {a: [[chrom, size] for chrom, size in zip(s.index, s)]
                      for a, s in sizes.items()}},
            f)
    for anchor, anno in zip(anchors, anno):
        print("collecting regional scores")
        a = ".".join(os.path.basename(anchor).split(".")[:-2])
        regcov_dict = pk_results.get_collapsed_regional_scores(
            anchor, {c: [[1, sizes[a][c]]] for c in sizes[a].index}
        )
        print("converting to bytes")
        scores = b"".join(
            int("".join(str(i) for i in s), 2).to_bytes(
                score_bitsize, byteorder="big", signed=False
            )
            for c in sizes[a].index
            for s in regcov_dict[c][(1, sizes[a][c])]
        )
        print("writing scoremap")
        for step in 1, LOWRES_STEP:
            with BgzfWriter(
                os.path.join(output_dir, "anchors", f"{a}.{step}.bgz"), "wb"
            ) as scores_out:
                scores_out.write(scores[::step] if step > 1 else scores)
            subprocess.run(
                (
                    "bgzip",
                    "-rI",
                    os.path.join(output_dir, "anchors", f"{a}.{step}.gzi"),
                    os.path.join(output_dir, "anchors", f"{a}.{step}.bgz"),
                )
            )
        print("formatting annotations")
        for type in "gene", "anno":
            bed = os.path.join(output_dir, "anno", f"{a}.{type}.bed")
            tbx = os.path.join(output_dir, "anno", f"{a}.{type}.bed.bgz")
            pybedtools.BedTool(gff2bed.convert(gff2bed.parse(anno))).saveas(
                bed
            ).sort().saveas(bed)
            pysam.tabix_compress(bed, tbx, True)
            pysam.tabix_index(tbx, True, 0, 1, 2, csi=True)



import urllib3
import shutil
import os
import os.path
import tarfile
import ftplib
import tempfile
import subprocess
from pankmer.env import EXAMPLE_DATA_URL, EXAMPLE_DATA_DIR
from pankmer.datasets import (
    SPOLYRHIZA_URL,
    SPOLYRHIZA_FILES,
    SOLANUM_URL_FORMAT,
    SOLANUM_IDS,
    ZEA_URL,
    ZEA_FILES,
    HSAPIENS_URL,
    HSAPIENS_AGC,
    HSAPIENS_AGC_BINARY_URL,
    HSAPIENS_IDS,
    ATHALIANA_URL_FORMAT,
    ATHALIANA_IDS,
    BSUBTILIS_DATA_FTP,
    BSUBTILIS_DATA_PATHS,
)


def download_example(dir: str = EXAMPLE_DATA_DIR, clade=None, n_samples: int = 1):
    """Download an example datatset. A. thaliana pseudo-genomes from 1001
    Genomes. These sequence data were produced by the Weigel laboratory at the
    Max Planck Institute for Developmental Biology.

    Parameters
    ----------
    dir : str
        Destination directory for example data
    clade
        If None, download small example dataset. if str, download publicly
        available genomes. Clade: max_samples. Spolyrhiza: 3, Bsubtilis: 164,
        Solanum: 46, Zea: 54, Hsapiens: 94, Athaliana: 1135
    n_samples : int
        Number of samples to download, must be less than max [1]
    """

    if clade == "Spolyrhiza":
        if n_samples > 3:
            raise RuntimeError("n_samples parameter must be <= 3 for Spolyrhiza")
        http = urllib3.PoolManager()
        with tempfile.TemporaryDirectory(dir=dir) as temp_dir:
            tar_file_path = os.path.join(temp_dir, os.path.basename(SPOLYRHIZA_URL))
            if os.path.exists(tar_file_path[:-7]):
                raise RuntimeError("destination already exists")
            with http.request("GET", SPOLYRHIZA_URL, preload_content=False) as r, open(
                tar_file_path, "wb"
            ) as out_file:
                shutil.copyfileobj(r, out_file)
            with tarfile.open(tar_file_path) as tar:
                tar.extractall(temp_dir)
            os.remove(tar_file_path)
            for fasta_file in SPOLYRHIZA_FILES[:n_samples]:
                shutil.move(
                    os.path.join(temp_dir, "Spolyrhiza_genomes", fasta_file),
                    os.path.join(dir, fasta_file),
                )
            shutil.rmtree(os.path.join(temp_dir, "Spolyrhiza_genomes"))

    elif clade == "Solanum":
        if n_samples > 46:
            raise RuntimeError("n_samples parameter must be <= 46 for Solanum")
        http = urllib3.PoolManager()
        for genome_id in SOLANUM_IDS[:n_samples]:
            fasta_url = SOLANUM_URL_FORMAT.format(genome_id)
            dest_file_path = os.path.join(dir, os.path.basename(fasta_url))
            if os.path.exists(dest_file_path):
                raise RuntimeError("destination already exists")
            with http.request("GET", fasta_url, preload_content=False) as r, open(
                dest_file_path, "wb"
            ) as out_file:
                shutil.copyfileobj(r, out_file)

    elif clade == "Zea":
        if n_samples > 54:
            raise RuntimeError("n_samples parameter must be <= 54 for Zea")
        http = urllib3.PoolManager()
        for fasta_file in ZEA_FILES[:n_samples]:
            fasta_url = os.path.join(ZEA_URL, fasta_file)
            dest_file_path = os.path.join(dir, os.path.basename(fasta_url))
            if os.path.exists(dest_file_path):
                raise RuntimeError("destination already exists")
            with http.request("GET", fasta_url, preload_content=False) as r, open(
                dest_file_path, "wb"
            ) as out_file:
                shutil.copyfileobj(r, out_file)

    elif clade == "Hsapiens":
        if n_samples > 94:
            raise RuntimeError("n_samples parameter must be <= 94 for Hsapiens")
        http = urllib3.PoolManager()
        with tempfile.TemporaryDirectory(dir=dir) as temp_dir:
            tar_file_path = os.path.join(
                temp_dir, os.path.basename(HSAPIENS_AGC_BINARY_URL)
            )
            agc_binary_path = os.path.join(temp_dir, "agc-1.1_x64-linux", "agc")
            agc_file_path = os.path.join(temp_dir, HSAPIENS_AGC)
            with http.request(
                "GET", HSAPIENS_AGC_BINARY_URL, preload_content=False
            ) as r, open(tar_file_path, "wb") as out_file:
                shutil.copyfileobj(r, out_file)
            with tarfile.open(tar_file_path) as tar:
                tar.extractall(temp_dir)
            with http.request("GET", HSAPIENS_URL, preload_content=False) as r, open(
                agc_file_path, "wb"
            ) as out_file:
                shutil.copyfileobj(r, out_file)
            for genome_id in HSAPIENS_IDS[:n_samples]:
                with open(os.path.join(dir, f"{genome_id}.fa"), "wb") as f:
                    subprocess.run(
                        (agc_binary_path, "getset", agc_file_path, genome_id), stdout=f
                    )

    elif clade == "Bsubtilis":
        if n_samples > 164:
            raise RuntimeError("n_samples parameter must be <= 164 for Bsubtillis")
        ftp = ftplib.FTP(BSUBTILIS_DATA_FTP)
        ftp.login()
        for ftp_path in BSUBTILIS_DATA_PATHS[:n_samples]:
            with open(os.path.join(dir, os.path.basename(ftp_path)), "wb") as f:
                ftp.retrbinary(f"RETR {ftp_path}", f.write)

    elif clade == "Athaliana":
        if n_samples > 1135:
            raise RuntimeError("n_samples parameter must be <= 1135 for Athaliana")
        http = urllib3.PoolManager()
        for pseudo in ATHALIANA_IDS[:n_samples]:
            fasta_url = ATHALIANA_URL_FORMAT.format(pseudo)
            dest_file_path = os.path.join(dir, os.path.basename(fasta_url))
            if os.path.exists(dest_file_path):
                raise RuntimeError("destination already exists")
            with http.request("GET", fasta_url, preload_content=False) as r, open(
                dest_file_path, "wb"
            ) as out_file:
                shutil.copyfileobj(r, out_file)

    elif isinstance(clade, str):
        raise RuntimeError("invalid clade")

    else:
        http = urllib3.PoolManager()
        tar_file_path = os.path.join(dir, os.path.basename(EXAMPLE_DATA_URL))
        if os.path.exists(tar_file_path[:-7]):
            raise RuntimeError("destination already exists")
        with http.request("GET", EXAMPLE_DATA_URL, preload_content=False) as r, open(
            tar_file_path, "wb"
        ) as out_file:
            shutil.copyfileobj(r, out_file)
        with tarfile.open(tar_file_path) as tar:
            tar.extractall(dir)
        os.remove(tar_file_path)
